__all__ = ["attn_layers"]

from tensorflow.keras import initializers, regularizers, constraints
from tensorflow.python.keras.engine import base_layer_utils
from tensorflow.keras.layers import Layer
from tensorflow import keras
from tensorflow.keras import layers
import tensorflow.keras.backend as K
from tensorflow.python.ops import math_ops
from tensorflow.python.ops import array_ops
from tensorflow.keras.layers import Dense, Lambda, dot, Activation, concatenate, Softmax
import tensorflow as tf
import math
import numpy as np

# A review of different attention mechanisms is given at following link
# https://lilianweng.github.io/lil-log/2018/06/24/attention-attention.html
# Raffel, SeqWeightedSelfAttention and HierarchicalAttention appear to be very much similar.

class AttentionRaffel(Layer):
    """
    Keras Layer that implements an Attention mechanism for temporal data.
    Supports Masking.
    Follows the work of Raffel et al., 2015 https://arxiv.org/pdf/1512.08756.pdf
    # Input shape
        3D tensor with shape: `(samples, steps, features)`.
    # Output shape
        2D tensor with shape: `(samples, features)`.
    :param kwargs:
    Just put it on top of an RNN Layer (GRU/LSTM/SimpleRNN) with return_sequences=True.
    The dimensions are inferred based on the output shape of the RNN.
    Example:
        model.add(LSTM(64, return_sequences=True))
        model.add(Attention())
    """
    def __init__(self, step_dim,
                 W_regularizer=None, b_regularizer=None,
                 W_constraint=None, b_constraint=None,
                 bias=True, **kwargs):
        self.supports_masking = True
        self.init = initializers.get('glorot_uniform')

        self.W_regularizer = regularizers.get(W_regularizer)
        self.b_regularizer = regularizers.get(b_regularizer)

        self.W_constraint = constraints.get(W_constraint)
        self.b_constraint = constraints.get(b_constraint)

        self.bias = bias
        self.step_dim = step_dim
        self.features_dim = 0
        super(AttentionRaffel, self).__init__(**kwargs)

    def build(self, input_shape):
        assert len(input_shape) == 3

        self.W = self.add_weight(shape=(input_shape[-1], ),
                                 initializer=self.init,
                                 name='{}_W'.format(self.name),
                                 regularizer=self.W_regularizer,
                                 constraint=self.W_constraint)
        self.features_dim = input_shape[-1]

        if self.bias:
            self.b = self.add_weight(shape=(input_shape[1],),
                                     initializer='zero',
                                     name='{}_b'.format(self.name),
                                     regularizer=self.b_regularizer,
                                     constraint=self.b_constraint)
        else:
            self.b = None

        self.built = True

    def compute_mask(self, input, input_mask=None):
        return None

    def __call__(self, x, mask=None):
        # TODO different result when using call()
        if not self.built:
            self._maybe_build(x)

        features_dim = self.features_dim
        step_dim = self.step_dim

        eij = K.reshape(K.dot(K.reshape(x, (-1, features_dim)),
                        K.reshape(self.W, (features_dim, 1))), (-1, step_dim))

        if self.bias:
            eij += self.b

        eij = tf.tanh(eij, name="AttentionRaffel_tanh")

        a = math_ops.exp(eij, name="AttentionRaffel_exp")

        if mask is not None:
            a *= K.cast(mask, K.floatx())

        a = tf.divide(a, K.cast(math_ops.reduce_sum(a, axis=1, keepdims=True,
                                                    name="AttentionRaffel_sum") + K.epsilon(), K.floatx()),
                      name="AttentionRaffel_a")

        a = array_ops.expand_dims(a, axis=-1, name="AttentionRaffel_expand_dims")
        weighted_input = x * a
        return K.sum(weighted_input, axis=1)

    def compute_output_shape(self, input_shape):
        return input_shape[0],  self.features_dim


class SelfAttention(Layer):

    def __init__(self, attn_units=128, attn_activation='tanh', context='', **kwargs):
        self.attn_units = attn_units
        self.attn_activation = attn_activation
        self.context = context
        super().__init__(**kwargs)

    def __call__(self, hidden_states):
        """
        SelfAttention is originally proposed by Cheng et al., 2016 https://arxiv.org/pdf/1601.06733.pdf
        Here using the implementation of Philipperemy from
        https://github.com/philipperemy/keras-attention-mechanism/blob/master/attention/attention.py with modification
        that `attn_units` and `attn_activation` attributes can be changed. The default values of these attributes are same
        as used by the auther. However, there is another implementation of SelfAttention at
        https://github.com/CyberZHG/keras-self-attention/blob/master/keras_self_attention/seq_self_attention.py
        but the author have cited a different paper i.e. Zheng et al., 2018 https://arxiv.org/pdf/1806.01264.pdf and
        named it as additive attention.
        A useful discussion about this (in this class) implementation can be found at
        https://github.com/philipperemy/keras-attention-mechanism/issues/14
        Many-to-one attention mechanism for Keras.
        @param hidden_states: 3D tensor with shape (batch_size, time_steps, input_dim).
        @return: 2D tensor with shape (batch_size, 128)
        @author: felixhao28.
        The original code which has here been modified had Apache Licence 2.0.
        """
        hidden_size = int(hidden_states.shape[2])
        # Inside dense layer
        #              hidden_states            dot               W            =>           score_first_part
        # (batch_size, time_steps, hidden_size) dot (hidden_size, hidden_size) => (batch_size, time_steps, hidden_size)
        # W is the trainable weight matrix of attention Luong's multiplicative style score
        score_first_part = Dense(hidden_size, use_bias=False, name='attention_score_vec' + self.context)(hidden_states)
        #            score_first_part           dot        last_hidden_state     => attention_weights
        # (batch_size, time_steps, hidden_size) dot   (batch_size, hidden_size)  => (batch_size, time_steps)
        h_t = Lambda(lambda x: x[:, -1, :], output_shape=(hidden_size,), name='last_hidden_state' + self.context)(hidden_states)
        score = dot([score_first_part, h_t], [2, 1], name='attention_score' + self.context)
        attention_weights = Activation('softmax', name='attention_weight' + self.context)(score)
        # (batch_size, time_steps, hidden_size) dot (batch_size, time_steps) => (batch_size, hidden_size)
        context_vector = dot([hidden_states, attention_weights], [1, 1], name='context_vector' + self.context)
        pre_activation = concatenate([context_vector, h_t], name='attention_output' + self.context)
        attention_vector = Dense(self.attn_units, use_bias=False, activation=self.attn_activation, name='attention_vector' + self.context)(pre_activation)
        return attention_vector


class BahdanauAttention(Layer):
    """
    Also known as Additive attention.
    https://github.com/thushv89/attention_keras/blob/master/src/layers/attention.py
    This can be implemented in encoder-decoder model as shown here.
    https://github.com/thushv89/attention_keras/blob/master/src/examples/nmt/model.py

    This class implements Bahdanau attention (https://arxiv.org/pdf/1409.0473.pdf).

    There are three sets of weights introduced W_a, U_a, and V_a

    The original code had following MIT licence
     ----------------------------------------------------------
     MIT License

    Copyright (c) 2019 Thushan Ganegedara

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.

     ----------------------------------------------------------
     """

    def __init__(self, **kwargs):

        super(BahdanauAttention, self).__init__(**kwargs)

    def build(self, input_shape):

        assert isinstance(input_shape, list)

        # Create a trainable weight variable for this layer.


        self.W_a = self.add_weight(name='W_a',
                                   shape=tf.TensorShape((input_shape[0][2], input_shape[0][2])),
                                   initializer='uniform',
                                   trainable=True)

        self.U_a = self.add_weight(name='U_a',
                                   shape=tf.TensorShape((input_shape[1][2], input_shape[0][2])),
                                   initializer='uniform',
                                   trainable=True)

        self.V_a = self.add_weight(name='V_a',
                                   shape=tf.TensorShape((input_shape[0][2], 1)),
                                   initializer='uniform',
                                   trainable=True)

        super(BahdanauAttention, self).build(input_shape)


    def __call__(self, inputs, verbose=False):
        """
        inputs: [encoder_output_sequence, decoder_output_sequence]
        """

        if not self.built:
            self._maybe_build(inputs)

        assert type(inputs) == list

        encoder_out_seq, decoder_out_seq = inputs

        if verbose:
            print('encoder_out_seq>', encoder_out_seq.shape)
            print('decoder_out_seq>', decoder_out_seq.shape)

        def energy_step(inputs, states):

            """ Step function for computing energy for a single decoder state """

            assert_msg = "States must be a list. However states {} is of type {}".format(states, type(states))

            assert isinstance(states, list) or isinstance(states, tuple), assert_msg

            """ Some parameters required for shaping tensors"""

            en_seq_len, en_hidden = encoder_out_seq.shape[1], encoder_out_seq.shape[2]

            de_hidden = inputs.shape[-1]

            """ Computing S.Wa where S=[s0, s1, ..., si]"""
            # batch_size*en_seq_len, latent_dim
            reshaped_enc_outputs = K.reshape(encoder_out_seq, (-1, en_hidden))

            # batch_size*en_seq_len, latent_dim

            W_a_dot_s = K.reshape(K.dot(reshaped_enc_outputs, self.W_a), (-1, en_seq_len, en_hidden))

            if verbose:
                print('wa.s>',W_a_dot_s.shape)

            """ Computing hj.Ua """
            # batch_size, 1, latent_dim
            U_a_dot_h = K.expand_dims(K.dot(inputs, self.U_a), 1)

            if verbose:
                print('Ua.h>',U_a_dot_h.shape)

            """ tanh(S.Wa + hj.Ua) """
            # batch_size*en_seq_len, latent_dim
            reshaped_Ws_plus_Uh = K.tanh(K.reshape(W_a_dot_s + U_a_dot_h, (-1, en_hidden)))

            if verbose:
                print('Ws+Uh>', reshaped_Ws_plus_Uh.shape)

            """ softmax(va.tanh(S.Wa + hj.Ua)) """
            # batch_size, en_seq_len
            e_i = K.reshape(K.dot(reshaped_Ws_plus_Uh, self.V_a), (-1, en_seq_len))

            # batch_size, en_seq_len
            e_i = K.softmax(e_i, name='softmax')

            if verbose:
                print('ei>', e_i.shape)

            return e_i, [e_i]

        def context_step(inputs, states):
            """ Step function for computing ci using ei """

            # batch_size, hidden_size
            c_i = K.sum(encoder_out_seq * K.expand_dims(inputs, -1), axis=1)

            if verbose:
                print('ci>', c_i.shape)

            return c_i, [c_i]

        def create_inital_state(inputs, hidden_size):

            # We are not using initial states, but need to pass something to K.rnn funciton

            fake_state = K.zeros_like(inputs)  # <= (batch_size, enc_seq_len, latent_dim)

            fake_state = K.sum(fake_state, axis=[1, 2])  # <= (batch_size)

            fake_state = K.expand_dims(fake_state)  # <= (batch_size, 1)

            fake_state = K.tile(fake_state, [1, hidden_size])  # <= (batch_size, latent_dim)

            return fake_state

        fake_state_c = create_inital_state(encoder_out_seq, encoder_out_seq.shape[-1])

        fake_state_e = create_inital_state(encoder_out_seq, encoder_out_seq.shape[1])  # <= (batch_size, enc_seq_len, latent_dim)


        """ Computing energy outputs """
        # e_outputs => (batch_size, de_seq_len, en_seq_len)
        last_out, e_outputs, _ = K.rnn(
            energy_step, decoder_out_seq, [fake_state_e],
        )

        """ Computing context vectors """
        last_out, c_outputs, _ = K.rnn(
            context_step, e_outputs, [fake_state_c],
        )
        return c_outputs, e_outputs

    def compute_output_shape(self, input_shape):

        """ Outputs produced by the layer """
        return [
            tf.TensorShape((input_shape[1][0], input_shape[1][1], input_shape[1][2])),
            tf.TensorShape((input_shape[1][0], input_shape[1][1], input_shape[0][1]))
        ]


def dot_product(x, kernel):
    """
    Wrapper for dot product operation, in order to be compatible with both
    Theano and Tensorflow
    Args:
        x (): input
        kernel (): weights
    Returns:
    """
    if K.backend() == 'tensorflow':
        return K.squeeze(K.dot(x, K.expand_dims(kernel)), axis=-1)
    else:
        return K.dot(x, kernel)


class HierarchicalAttention(Layer):
    """
    Used from https://gist.github.com/cbaziotis/7ef97ccf71cbc14366835198c09809d2
    Attention operation, with a context/query vector, for temporal data.
    Supports Masking.
    Follows the work of Yang et al., 2016 [https://www.cs.cmu.edu/~./hovy/papers/16HLT-hierarchical-attention-networks.pdf]
    "Hierarchical Attention Networks for Document Classification"
    by using a context vector to assist the attention
    # Input shape
        3D tensor with shape: `(samples, steps, features)`.
    # Output shape
        2D tensor with shape: `(samples, features)`.
    How to use:
    Just put it on top of an RNN Layer (GRU/LSTM/SimpleRNN) with return_sequences=True.
    The dimensions are inferred based on the output shape of the RNN.
    Note: The layer has been tested with Keras 2.0.6
    Example:
        model.add(LSTM(64, return_sequences=True))
        model.add(AttentionWithContext())
        # next add a Dense layer (for classification/regression) or whatever...
    """

    def __init__(self,
                 W_regularizer=None, u_regularizer=None, b_regularizer=None,
                 W_constraint=None, u_constraint=None, b_constraint=None,
                 bias=True, **kwargs):

        self.supports_masking = True
        self.init = initializers.get('glorot_uniform')

        self.W_regularizer = regularizers.get(W_regularizer)
        self.u_regularizer = regularizers.get(u_regularizer)
        self.b_regularizer = regularizers.get(b_regularizer)

        self.W_constraint = constraints.get(W_constraint)
        self.u_constraint = constraints.get(u_constraint)
        self.b_constraint = constraints.get(b_constraint)

        self.bias = bias
        super(HierarchicalAttention, self).__init__(**kwargs)

    def build(self, input_shape):
        assert len(input_shape) == 3

        self.W = self.add_weight(shape=(input_shape[-1], input_shape[-1],),
                                 initializer=self.init,
                                 name='{}_W'.format(self.name),
                                 regularizer=self.W_regularizer,
                                 constraint=self.W_constraint)
        if self.bias:
            self.b = self.add_weight(shape=(input_shape[-1],),
                                     initializer='zero',
                                     name='{}_b'.format(self.name),
                                     regularizer=self.b_regularizer,
                                     constraint=self.b_constraint)

        self.u = self.add_weight(shape=(input_shape[-1],),
                                 initializer=self.init,
                                 name='{}_u'.format(self.name),
                                 regularizer=self.u_regularizer,
                                 constraint=self.u_constraint)

        super(HierarchicalAttention, self).build(input_shape)

    def compute_mask(self, input, input_mask=None):
        # do not pass the mask to the next layers
        return None

    def __call__(self, x, mask=None):
        # TODO different result when using call()
        if not self.built:
            self._maybe_build(x)

        uit = dot_product(x, self.W)  # eq 5 in paper

        if self.bias:
            uit += self.b

        uit = K.tanh(uit)  # eq 5 in paper
        ait = dot_product(uit, self.u)

        a = K.exp(ait)

        # apply mask after the exp. will be re-normalized next
        if mask is not None:
            # Cast the mask to floatX to avoid float64 upcasting in theano
            a *= K.cast(mask, K.floatx())

        # in some cases especially in the early stages of training the sum may be almost zero
        # and this results in NaN's. A workaround is to add a very small positive number ε to the sum.
        # a /= K.cast(K.sum(a, axis=1, keepdims=True), K.floatx())
        # eq 6 in paper
        a /= math_ops.cast(math_ops.reduce_sum(a, axis=1, keepdims=True, name="HA_sum") + K.epsilon(), K.floatx(),
                           name="HA_cast")

        a = array_ops.expand_dims(a, axis=-1, name="HA_expand_dims")
        weighted_input = tf.math.multiply(x,a,name="HA_weighted_input")  #x * a
        return K.sum(weighted_input, axis=1)  # eq 7 in paper

    def compute_output_shape(self, input_shape):
        return input_shape[0], input_shape[-1]


class SeqSelfAttention(Layer):

    ATTENTION_TYPE_ADD = 'additive'
    ATTENTION_TYPE_MUL = 'multiplicative'

    def __init__(self,
                 units=32,
                 attention_width=None,
                 attention_type=ATTENTION_TYPE_ADD,
                 return_attention=False,
                 history_only=False,
                 kernel_initializer='glorot_normal',
                 bias_initializer='zeros',
                 kernel_regularizer=None,
                 bias_regularizer=None,
                 kernel_constraint=None,
                 bias_constraint=None,
                 use_additive_bias=True,
                 use_attention_bias=True,
                 attention_activation=None,
                 attention_regularizer_weight=0.0,
                 **kwargs):
        """
        using implementation of
        https://github.com/CyberZHG/keras-self-attention/blob/master/keras_self_attention/seq_self_attention.py
        Layer initialization.

        For additive attention, see: https://arxiv.org/pdf/1806.01264.pdf

        :param units: The dimension of the vectors that used to calculate the attention weights.
        :param attention_width: The width of local attention.
        :param attention_type: 'additive' or 'multiplicative'.
        :param return_attention: Whether to return the attention weights for visualization.
        :param history_only: Only use historical pieces of data.
        :param kernel_initializer: The initializer for weight matrices.
        :param bias_initializer: The initializer for biases.
        :param kernel_regularizer: The regularization for weight matrices.
        :param bias_regularizer: The regularization for biases.
        :param kernel_constraint: The constraint for weight matrices.
        :param bias_constraint: The constraint for biases.
        :param use_additive_bias: Whether to use bias while calculating the relevance of inputs features
                                  in additive mode.
        :param use_attention_bias: Whether to use bias while calculating the weights of attention.
        :param attention_activation: The activation used for calculating the weights of attention.
        :param attention_regularizer_weight: The weights of attention regularizer.
        :param kwargs: Parameters for parent class.

        The original code which has here been slightly modified came with following licence.
        ----------------------------------
        MIT License

        Copyright (c) 2018 PoW

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.
        -----------------------------------------------------------------------
        """
        super(SeqSelfAttention, self).__init__(**kwargs)
        self.supports_masking = True
        self.units = units
        self.attention_width = attention_width
        self.attention_type = attention_type
        self.return_attention = return_attention
        self.history_only = history_only
        if history_only and attention_width is None:
            self.attention_width = int(1e9)

        self.use_additive_bias = use_additive_bias
        self.use_attention_bias = use_attention_bias
        self.kernel_initializer = keras.initializers.get(kernel_initializer)
        self.bias_initializer = keras.initializers.get(bias_initializer)
        self.kernel_regularizer = keras.regularizers.get(kernel_regularizer)
        self.bias_regularizer = keras.regularizers.get(bias_regularizer)
        self.kernel_constraint = keras.constraints.get(kernel_constraint)
        self.bias_constraint = keras.constraints.get(bias_constraint)
        self.attention_activation = keras.activations.get(attention_activation)
        self.attention_regularizer_weight = attention_regularizer_weight
        self._backend = keras.backend.backend()

        if attention_type.upper().startswith('ADD'):
            self.Wx, self.Wt, self.bh = None, None, None
            self.Wa, self.ba = None, None
        elif attention_type.upper().startswith('MUL'):
            self.Wa, self.ba = None, None
        else:
            raise NotImplementedError('No implementation for attention type : ' + attention_type)

    def get_config(self):
        config = {
            'units': self.units,
            'attention_width': self.attention_width,
            'attention_type': self.attention_type,
            'return_attention': self.return_attention,
            'history_only': self.history_only,
            'use_additive_bias': self.use_additive_bias,
            'use_attention_bias': self.use_attention_bias,
            'kernel_initializer': keras.initializers.serialize(self.kernel_initializer),
            'bias_initializer': keras.initializers.serialize(self.bias_initializer),
            'kernel_regularizer': keras.regularizers.serialize(self.kernel_regularizer),
            'bias_regularizer': keras.regularizers.serialize(self.bias_regularizer),
            'kernel_constraint': keras.constraints.serialize(self.kernel_constraint),
            'bias_constraint': keras.constraints.serialize(self.bias_constraint),
            'attention_activation': keras.activations.serialize(self.attention_activation),
            'attention_regularizer_weight': self.attention_regularizer_weight,
        }
        base_config = super(SeqSelfAttention, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))

    def build(self, input_shape):
        if self.attention_type.upper().startswith('ADD'):
            self._build_additive_attention(input_shape)
        elif self.attention_type.upper().startswith('MUL'):
            self._build_multiplicative_attention(input_shape)
        super(SeqSelfAttention, self).build(input_shape)

    def _build_additive_attention(self, input_shape):
        feature_dim = int(input_shape[2])

        self.Wt = self.add_weight(shape=(feature_dim, self.units),
                                  name='{}_Add_Wt'.format(self.name),
                                  initializer=self.kernel_initializer,
                                  regularizer=self.kernel_regularizer,
                                  constraint=self.kernel_constraint)
        self.Wx = self.add_weight(shape=(feature_dim, self.units),
                                  name='{}_Add_Wx'.format(self.name),
                                  initializer=self.kernel_initializer,
                                  regularizer=self.kernel_regularizer,
                                  constraint=self.kernel_constraint)
        if self.use_additive_bias:
            self.bh = self.add_weight(shape=(self.units,),
                                      name='{}_Add_bh'.format(self.name),
                                      initializer=self.bias_initializer,
                                      regularizer=self.bias_regularizer,
                                      constraint=self.bias_constraint)

        self.Wa = self.add_weight(shape=(self.units, 1),
                                  name='{}_Add_Wa'.format(self.name),
                                  initializer=self.kernel_initializer,
                                  regularizer=self.kernel_regularizer,
                                  constraint=self.kernel_constraint)
        if self.use_attention_bias:
            self.ba = self.add_weight(shape=(1,),
                                      name='{}_Add_ba'.format(self.name),
                                      initializer=self.bias_initializer,
                                      regularizer=self.bias_regularizer,
                                      constraint=self.bias_constraint)

    def _build_multiplicative_attention(self, input_shape):
        feature_dim = int(input_shape[2])

        self.Wa = self.add_weight(shape=(feature_dim, feature_dim),
                                  name='{}_Mul_Wa'.format(self.name),
                                  initializer=self.kernel_initializer,
                                  regularizer=self.kernel_regularizer,
                                  constraint=self.kernel_constraint)
        if self.use_attention_bias:
            self.ba = self.add_weight(shape=(1,),
                                      name='{}_Mul_ba'.format(self.name),
                                      initializer=self.bias_initializer,
                                      regularizer=self.bias_regularizer,
                                      constraint=self.bias_constraint)

    def __call__(self, inputs, mask=None, **kwargs):
        # TODO different result when using call()
        if not self.built:
            self._maybe_build(inputs)

        input_len = K.shape(inputs)[1]

        if self.attention_type.upper().startswith('ADD'):
            e = self._call_additive_emission(inputs)
        else:
            e = self._call_multiplicative_emission(inputs)

        if self.attention_activation is not None:
            e = self.attention_activation(e)
        if self.attention_width is not None:
            if self.history_only:
                lower = K.arange(0, input_len) - (self.attention_width - 1)
            else:
                lower = K.arange(0, input_len) - self.attention_width // 2
            lower = K.expand_dims(lower, axis=-1)
            upper = lower + self.attention_width
            indices = K.expand_dims(K.arange(0, input_len), axis=0)
            e -= 10000.0 * (1.0 - K.cast(lower <= indices, K.floatx()) * K.cast(indices < upper, K.floatx()))
        if mask is not None:
            mask = K.expand_dims(K.cast(mask, K.floatx()), axis=-1)
            e -= 10000.0 * ((1.0 - mask) * (1.0 - K.permute_dimensions(mask, (0, 2, 1))))

        # a_{t} = \text{softmax}(e_t)
        a = Softmax(axis=-1, name='SeqSelfAttention_Softmax')(e)

        # l_t = \sum_{t'} a_{t, t'} x_{t'}
        v = K.batch_dot(a, inputs)
        if self.attention_regularizer_weight > 0.0:
            self.add_loss(self._attention_regularizer(a))

        if self.return_attention:
            return [v, a]
        return v

    def _call_additive_emission(self, inputs):
        input_shape = K.shape(inputs)
        batch_size, input_len = input_shape[0], input_shape[1]

        # h_{t, t'} = \tanh(x_t^T W_t + x_{t'}^T W_x + b_h)
        q = K.expand_dims(K.dot(inputs, self.Wt), 2)
        k = K.expand_dims(K.dot(inputs, self.Wx), 1)
        if self.use_additive_bias:
            h = K.tanh(q + k + self.bh)
        else:
            h = K.tanh(q + k)

        # e_{t, t'} = W_a h_{t, t'} + b_a
        if self.use_attention_bias:
            e = K.reshape(K.dot(h, self.Wa) + self.ba, (batch_size, input_len, input_len))
        else:
            e = K.reshape(K.dot(h, self.Wa), (batch_size, input_len, input_len))
        return e

    def _call_multiplicative_emission(self, inputs):
        # e_{t, t'} = x_t^T W_a x_{t'} + b_a
        e = K.batch_dot(K.dot(inputs, self.Wa), K.permute_dimensions(inputs, (0, 2, 1)))
        if self.use_attention_bias:
            e += self.ba[0]
        return e

    def compute_output_shape(self, input_shape):
        output_shape = input_shape
        if self.return_attention:
            attention_shape = (input_shape[0], output_shape[1], input_shape[1])
            return [output_shape, attention_shape]
        return output_shape

    def compute_mask(self, inputs, mask=None):
        if self.return_attention:
            return [mask, None]
        return mask

    def _attention_regularizer(self, attention):
        batch_size = K.cast(K.shape(attention)[0], K.floatx())
        input_len = K.shape(attention)[-1]
        indices = K.expand_dims(K.arange(0, input_len), axis=0)
        diagonal = K.expand_dims(K.arange(0, input_len), axis=-1)
        eye = K.cast(K.equal(indices, diagonal), K.floatx())
        return self.attention_regularizer_weight * K.sum(K.square(K.batch_dot(
            attention,
            K.permute_dimensions(attention, (0, 2, 1))) - eye)) / batch_size

    @staticmethod
    def get_custom_objects():
        return {'SeqSelfAttention': SeqSelfAttention}


class SeqWeightedAttention(keras.layers.Layer):
    r"""Y = \text{softmax}(XW + b) X
    See: https://arxiv.org/pdf/1708.00524.pdf
    using implementation of https://github.com/CyberZHG/keras-self-attention/blob/master/keras_self_attention/seq_weighted_attention.py
            The original code which has here been slightly modified came with following licence.
        -----------------------------------------------------------------------
        MIT License

        Copyright (c) 2018 PoW

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.
        -----------------------------------------------------------------------

    """

    def __init__(self, use_bias=True, return_attention=False, **kwargs):
        super(SeqWeightedAttention, self).__init__(**kwargs)
        self.supports_masking = True
        self.use_bias = use_bias
        self.return_attention = return_attention
        self.W, self.b = None, None

    def get_config(self):
        config = {
            'use_bias': self.use_bias,
            'return_attention': self.return_attention,
        }
        base_config = super(SeqWeightedAttention, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))

    def build(self, input_shape):
        self.W = self.add_weight(shape=(int(input_shape[2]), 1),
                                 name='{}_W'.format(self.name),
                                 initializer=keras.initializers.get('uniform'))
        if self.use_bias:
            self.b = self.add_weight(shape=(1,),
                                     name='{}_b'.format(self.name),
                                     initializer=keras.initializers.get('zeros'))
        super(SeqWeightedAttention, self).build(input_shape)

    def __call__(self, x, mask=None):
        # TODO different result when using call()
        if not self.built:
            self._maybe_build(x)

        logits = K.dot(x, self.W)
        if self.use_bias:
            logits += self.b
        x_shape = K.shape(x)
        logits = K.reshape(logits, (x_shape[0], x_shape[1]))
        if mask is not None:
            mask = K.cast(mask, K.floatx())
            logits -= 10000.0 * (1.0 - mask)
        ai = math_ops.exp(logits - K.max(logits, axis=-1, keepdims=True), name="SeqWeightedAttention_exp")
        #att_weights = ai / (K.sum(ai, axis=1, keepdims=True) + K.epsilon())
        att_weights = tf.math.divide(ai, (math_ops.reduce_sum(ai, axis=1, keepdims=True,
                                                              name="SeqWeightedAttention_sum") + K.epsilon()),
                                     name="SeqWeightedAttention_weights")
        weighted_input = x * K.expand_dims(att_weights)
        result = K.sum(weighted_input, axis=1)
        if self.return_attention:
            return [result, att_weights]
        return result

    def compute_output_shape(self, input_shape):
        output_len = input_shape[2]
        if self.return_attention:
            return [(input_shape[0], output_len), (input_shape[0], input_shape[1])]
        return input_shape[0], output_len

    def compute_mask(self, _, input_mask=None):
        if self.return_attention:
            return [None, None]
        return None

    @staticmethod
    def get_custom_objects():
        return {'SeqWeightedAttention': SeqWeightedAttention}


class SnailAttention(Layer):
    """
    Based on work of Mishra et al., 2018 https://openreview.net/pdf?id=B1DmUzWAW
    Adopting code from https://github.com/philipperemy/keras-snail-attention/blob/master/attention.py
    """

    def __init__(self, dims, k_size, v_size, seq_len=None, **kwargs):
        self.k_size = k_size
        self.seq_len = seq_len
        self.v_size = v_size
        self.dims = dims
        self.sqrt_k = math.sqrt(k_size)
        self.keys_fc = None
        self.queries_fc = None
        self.values_fc = None
        super(SnailAttention, self).__init__(**kwargs)


    def build(self, input_shape):
        # https://stackoverflow.com/questions/54194724/how-to-use-keras-layers-in-custom-keras-layer
        self.keys_fc = Dense(self.k_size,  name="Keys_SnailAttn")
        self.keys_fc.build((None, self.dims))
        self._trainable_weights.extend(self.keys_fc.trainable_weights)

        self.queries_fc = Dense(self.k_size, name="Queries_SnailAttn")
        self.queries_fc.build((None, self.dims))
        self._trainable_weights.extend(self.queries_fc.trainable_weights)

        self.values_fc = Dense(self.v_size,  name="Values_SnailAttn")
        self.values_fc.build((None, self.dims))
        self._trainable_weights.extend(self.values_fc.trainable_weights)
        #super(SnailAttention, self).__init__(**kwargs)

    def __call__(self, inputs, **kwargs):

        if not self.built:
            self._maybe_build(inputs)

        # check that the implementation matches exactly py torch.
        keys = self.keys_fc(inputs)
        queries = self.queries_fc(inputs)
        values = self.values_fc(inputs)
        logits = K.batch_dot(queries, K.permute_dimensions(keys, (0, 2, 1)))
        mask = K.ones_like(logits) * np.triu((-np.inf) * np.ones(logits.shape.as_list()[1:]), k=1)
        logits = mask + logits
        probs = Softmax(axis=-1, name="Softmax_SnailAttn")(logits / self.sqrt_k)
        read = K.batch_dot(probs, values)
        output = K.concatenate([inputs, read], axis=-1)
        return output

    def compute_output_shape(self, input_shape):
        output_shape = list(input_shape)
        output_shape[-1] += self.v_size
        return tuple(output_shape)


def regularized_padded_conv(conv_dim, *args, **kwargs):
    if conv_dim == "1d":
        return layers.Conv1D(*args, **kwargs, padding='same', use_bias=False,
                             kernel_initializer='he_normal',
                             kernel_regularizer=regularizers.l2(5e-4))
    elif conv_dim == "2d":
        return layers.Conv2D(*args, **kwargs, padding='same', use_bias=False,
                             kernel_initializer='he_normal',
                             kernel_regularizer=regularizers.l2(5e-4))
    else:
        raise ValueError(f"conv_dim must be either 1d or 2d but it is {conv_dim}")

class ChannelAttention(layers.Layer):
    """Code adopted from https://github.com/zhangkaifang/CBAM-TensorFlow2.0.
    This feature attention generates time step context descriptors self.avg and self. max by using both average and
    max pooling operations along the time step axis and then fowards to a shared multi-layer perception (MLP) to produce
    the feature (channel) attention map."""
    def __init__(self, conv_dim, in_planes, ratio=16, **kwargs):

        if conv_dim not in ["1d", "2d"]:
            raise ValueError(f" conv_dim must be either 1d or 2d but it is {conv_dim}")

        super(ChannelAttention, self).__init__(**kwargs)
        if conv_dim == "1d":
            self.axis = (1,)
            self.avg= layers.GlobalAveragePooling1D()
            self.max= layers.GlobalMaxPooling1D()
            self.conv1 = layers.Conv1D(in_planes//ratio, kernel_size=1, strides=1, padding='same',
                                       kernel_regularizer=regularizers.l2(5e-4),
                                       use_bias=True, activation=tf.nn.relu, name='channel_attn1')
            self.conv2 = layers.Conv1D(in_planes, kernel_size=1, strides=1, padding='same',
                                       kernel_regularizer=regularizers.l2(5e-4),
                                       use_bias=True, name='channel_attn2')
        elif conv_dim == "2d":
            self.axis = (1,1)
            self.avg= layers.GlobalAveragePooling2D()
            self.max= layers.GlobalMaxPooling2D()
            self.conv1 = layers.Conv2D(in_planes//ratio, kernel_size=1, strides=1, padding='same',
                                       kernel_regularizer=regularizers.l2(5e-4),
                                       use_bias=True, activation=tf.nn.relu, name='channel_attn1')
            self.conv2 = layers.Conv2D(in_planes, kernel_size=1, strides=1, padding='same',
                                       kernel_regularizer=regularizers.l2(5e-4),
                                       use_bias=True, name='channe2_attn1')

    def __call__(self, inputs, *args):
        avg = self.avg(inputs)  # [256, 32, 32, 64] -> [256, 64]
        max_pool = self.max(inputs) # [256, 32, 32, 64] -> [256, 64]
        avg = layers.Reshape((*self.axis, avg.shape[1]))(avg)   # shape (None, 1, 1 feature)  # [256, 1, 1, 64]
        max_pool = layers.Reshape((*self.axis, max_pool.shape[1]))(max_pool)   # shape (None, 1, 1 feature)  # [256, 1, 1, 64]
        avg_out = self.conv2(self.conv1(avg))  # [256, 1, 1, 64] -> [256, 1, 1, 4] -> [256, 1, 1, 64]
        max_out = self.conv2(self.conv1(max_pool))  # [256, 1, 1, 64] -> [256, 1, 1, 4] -> [256, 1, 1, 64]
        out = avg_out + max_out  # [256, 1, 1, 64]
        out = tf.nn.sigmoid(out, name="ChannelAttention_sigmoid")  # [256, 1, 1, 64]

        return out


class SpatialAttention(layers.Layer):
    """Code adopted from https://github.com/zhangkaifang/CBAM-TensorFlow2.0 .
    The time step (spatial) attention module generates a concatenated feature descriptor [F'Tavg;F'Tmax]∈R2×T by
    applying average pooling and max pooling along the feature axis, followed by a standard convolution layer.[1].

    Cheng, Y., Liu, Z., & Morimoto, Y. (2020). Attention-Based SeriesNet: An Attention-Based Hybrid Neural Network Model
    for Conditional Time Series Forecasting. Information, 11(6), 305.
    """
    def __init__(self, conv_dim,  kernel_size=7, **kwargs):

        if conv_dim not in ["1d", "2d"]:
            raise ValueError(f" conv_dim must be either 1d or 2d but it is {conv_dim}")

        super(SpatialAttention, self).__init__(**kwargs)
        if conv_dim == "1d":
            self.axis = 2
        elif conv_dim == "2d":
            self.axis = 3
        self.conv1 = regularized_padded_conv(conv_dim,
                                             1, kernel_size=kernel_size, strides=1, activation=tf.nn.sigmoid, name="spatial_attn")

    def __call__(self, inputs, *args):
        avg_out = tf.reduce_mean(inputs, axis=self.axis, name="SpatialAttention_mean")  # [256, 32, 32, 64] -> [256, 32, 32]
        max_out = tf.reduce_max(inputs, axis=self.axis, name="SpatialAttention_max")  # [256, 32, 32, 64] -> [256, 32, 32]
        out = tf.stack([avg_out, max_out], axis=self.axis, name="SpatialAttention_stack")             # concat。 -> [256, 32, 32, 2]
        out = self.conv1(out)  # -> [256, 32, 32, 1]

        return out

class attn_layers(object):

    AttentionRaffel = AttentionRaffel
    SelfAttention = SelfAttention
    SeqSelfAttention = SeqSelfAttention
    SnailAttention = SnailAttention
    SeqWeightedAttention = SeqWeightedAttention
    BahdanauAttention = BahdanauAttention
    HierarchicalAttention = HierarchicalAttention
    SpatialAttention = SpatialAttention
    ChannelAttention = ChannelAttention



try:
    import torch
    nn = torch.nn
except (ImportError, ModuleNotFoundError):
    torch, nn = None, None


class IMVTensorLSTM(torch.jit.ScriptModule):
    __constants__ = ["n_units", "input_dim"]

    def __init__(self, input_dim, output_dim, n_units, device, init_std=0.02):
        super().__init__()
        self.U_j = nn.Parameter(torch.randn(input_dim, 1, n_units) * init_std)
        self.U_i = nn.Parameter(torch.randn(input_dim, 1, n_units) * init_std)
        self.U_f = nn.Parameter(torch.randn(input_dim, 1, n_units) * init_std)
        self.U_o = nn.Parameter(torch.randn(input_dim, 1, n_units) * init_std)
        self.W_j = nn.Parameter(torch.randn(input_dim, n_units, n_units) * init_std)
        self.W_i = nn.Parameter(torch.randn(input_dim, n_units, n_units) * init_std)
        self.W_f = nn.Parameter(torch.randn(input_dim, n_units, n_units) * init_std)
        self.W_o = nn.Parameter(torch.randn(input_dim, n_units, n_units) * init_std)
        self.b_j = nn.Parameter(torch.randn(input_dim, n_units) * init_std)
        self.b_i = nn.Parameter(torch.randn(input_dim, n_units) * init_std)
        self.b_f = nn.Parameter(torch.randn(input_dim, n_units) * init_std)
        self.b_o = nn.Parameter(torch.randn(input_dim, n_units) * init_std)
        self.F_alpha_n = nn.Parameter(torch.randn(input_dim, n_units, 1) * init_std)
        self.F_alpha_n_b = nn.Parameter(torch.randn(input_dim, 1) * init_std)
        self.F_beta = nn.Linear(2 * n_units, 1)
        self.Phi = nn.Linear(2 * n_units, output_dim)
        self.n_units = n_units
        self.input_dim = input_dim
        self.device = device

    @torch.jit.script_method
    def forward(self, x):
        h_tilda_t = torch.zeros(x.shape[0], self.input_dim, self.n_units).to(self.device)
        c_tilda_t = torch.zeros(x.shape[0], self.input_dim, self.n_units).to(self.device)
        outputs = torch.jit.annotate(List[Tensor], [])
        for t in range(x.shape[1]):
            # eq 1
            j_tilda_t = torch.tanh(torch.einsum("bij,ijk->bik", h_tilda_t, self.W_j) +
                                   torch.einsum("bij,jik->bjk", x[:, t, :].unsqueeze(1), self.U_j) + self.b_j)
            # eq 5
            i_tilda_t = torch.sigmoid(torch.einsum("bij,ijk->bik", h_tilda_t, self.W_i) +
                                      torch.einsum("bij,jik->bjk", x[:, t, :].unsqueeze(1), self.U_i) + self.b_i)
            f_tilda_t = torch.sigmoid(torch.einsum("bij,ijk->bik", h_tilda_t, self.W_f) +
                                      torch.einsum("bij,jik->bjk", x[:, t, :].unsqueeze(1), self.U_f) + self.b_f)
            o_tilda_t = torch.sigmoid(torch.einsum("bij,ijk->bik", h_tilda_t, self.W_o) +
                                      torch.einsum("bij,jik->bjk", x[:, t, :].unsqueeze(1), self.U_o) + self.b_o)
            # eq 6
            c_tilda_t = c_tilda_t * f_tilda_t + i_tilda_t * j_tilda_t
            # eq 7
            h_tilda_t = (o_tilda_t * torch.tanh(c_tilda_t))
            outputs += [h_tilda_t]
        outputs = torch.stack(outputs)
        outputs = outputs.permute(1, 0, 2, 3)
        # eq 8
        alphas = torch.tanh(torch.einsum("btij,ijk->btik", outputs, self.F_alpha_n) + self.F_alpha_n_b)
        alphas = torch.exp(alphas)
        alphas = alphas / torch.sum(alphas, dim=1, keepdim=True)
        g_n = torch.sum(alphas * outputs, dim=1)
        hg = torch.cat([g_n, h_tilda_t], dim=2)
        mu = self.Phi(hg)
        betas = torch.tanh(self.F_beta(hg))
        betas = torch.exp(betas)
        betas = betas / torch.sum(betas, dim=1, keepdim=True)
        mean = torch.sum(betas * mu, dim=1)

        return mean, alphas, betas


class IMVFullLSTM(torch.jit.ScriptModule):
    __constants__ = ["n_units", "input_dim"]

    def __init__(self, input_dim, output_dim, n_units, device, init_std=0.02):
        super().__init__()
        self.U_j = nn.Parameter(torch.randn(input_dim, 1, n_units) * init_std)
        self.W_j = nn.Parameter(torch.randn(input_dim, n_units, n_units) * init_std)
        self.b_j = nn.Parameter(torch.randn(input_dim, n_units) * init_std)
        self.W_i = nn.Linear(input_dim * (n_units + 1), input_dim * n_units)
        self.W_f = nn.Linear(input_dim * (n_units + 1), input_dim * n_units)
        self.W_o = nn.Linear(input_dim * (n_units + 1), input_dim * n_units)
        self.F_alpha_n = nn.Parameter(torch.randn(input_dim, n_units, 1) * init_std)
        self.F_alpha_n_b = nn.Parameter(torch.randn(input_dim, 1) * init_std)
        self.F_beta = nn.Linear(2 * n_units, 1)
        self.Phi = nn.Linear(2 * n_units, output_dim)
        self.n_units = n_units
        self.input_dim = input_dim
        self.device = device

    @torch.jit.script_method
    def forward(self, x):
        h_tilda_t = torch.zeros(x.shape[0], self.input_dim, self.n_units).to(self.device)
        c_t = torch.zeros(x.shape[0], self.input_dim * self.n_units).to(self.device)
        outputs = torch.jit.annotate(List[Tensor], [])
        for t in range(x.shape[1]):
            # eq 1
            j_tilda_t = torch.tanh(torch.einsum("bij,ijk->bik", h_tilda_t, self.W_j) +
                                   torch.einsum("bij,jik->bjk", x[:, t, :].unsqueeze(1), self.U_j) + self.b_j)
            inp = torch.cat([x[:, t, :], h_tilda_t.view(h_tilda_t.shape[0], -1)], dim=1)
            # eq 2
            i_t = torch.sigmoid(self.W_i(inp))
            f_t = torch.sigmoid(self.W_f(inp))
            o_t = torch.sigmoid(self.W_o(inp))
            # eq 3
            c_t = c_t * f_t + i_t * j_tilda_t.view(j_tilda_t.shape[0], -1)
            # eq 4
            h_tilda_t = (o_t * torch.tanh(c_t)).view(h_tilda_t.shape[0], self.input_dim, self.n_units)
            outputs += [h_tilda_t]
        outputs = torch.stack(outputs)
        outputs = outputs.permute(1, 0, 2, 3)
        # eq 8
        alphas = torch.tanh(torch.einsum("btij,ijk->btik", outputs, self.F_alpha_n) + self.F_alpha_n_b)
        alphas = torch.exp(alphas)
        alphas = alphas / torch.sum(alphas, dim=1, keepdim=True)
        g_n = torch.sum(alphas * outputs, dim=1)
        hg = torch.cat([g_n, h_tilda_t], dim=2)
        mu = self.Phi(hg)
        betas = torch.tanh(self.F_beta(hg))
        betas = torch.exp(betas)
        betas = betas / torch.sum(betas, dim=1, keepdim=True)
        mean = torch.sum(betas * mu, dim=1)
        return mean, alphas, betas



import gc
import copy
import math
import importlib
from typing import Union

from SeqMetrics import RegressionMetrics, ClassificationMetrics

from .utils.utils import dateandtime_now, jsonize, MATRIC_TYPES, update_model_config, TrainTestSplit


DEFAULTS = {
    'r2': 1.0,
    'nse': 1.0,
    'r2_score': 1.0
}

class ModelOptimizerMixIn(object):

    def __init__(
            self,
            model,
            algorithm,
            num_iterations,
            process_results,
            prefix="hpo",
    ):
        self.model = model
        self.algorithm = algorithm
        self.num_iterations = num_iterations
        self.process_results = process_results
        self.prefix = prefix

    def fit(self, data=None,):

        if isinstance(data, tuple) or isinstance(data, list):
            assert len(data) == 2
            xy = True
        else:
            xy = False

        PREFIX = f"{self.prefix}_{dateandtime_now()}"
        self.iter = 0
        print("{:<15} {:<20}".format("Iteration No.",  "Validation Score"))

        hpo = importlib.import_module("ai4water.hyperopt")

        val_metric = self.model.val_metric
        metric_type = MATRIC_TYPES.get(val_metric, 'min')
        cross_validator = self.model.config['cross_validator']
        config = jsonize(self.model.config)
        SEED = config['seed']

        if self.model.mode == "classification":
            Metrics = ClassificationMetrics
        else:
            Metrics = RegressionMetrics

        def objective_fn(
                **suggestions,
        ):
           # we must not set the seed here to None
           # this will cause random splitting unpreproducible (if random splitting is applied)
            config['verbosity'] = -1
            config['prefix'] = PREFIX

            getattr(self, f'update')(config, suggestions)

            _model = self.model.from_config(
                config.copy(),
                make_new_path=True,
            )

            if cross_validator is None:

                if xy:  # todo, it is better to split data outside objective_fn
                    splitter = TrainTestSplit(seed=SEED, 
                        test_fraction=config['val_fraction'])

                    if config['split_random']:
                        # for reproducibility, we should use SEED so that at everay optimization
                        # iteration, we split the data in the same way
                        train_x, test_x, train_y, test_y = splitter.split_by_random(*data)
                    else:
                        train_x, test_x, train_y, test_y = splitter.split_by_slicing(*data)

                    _model.fit(x=train_x, y=train_y)
                    p = _model.predict(test_x)
                else:
                    _model.fit(data=data)
                    test_y, p = _model.predict(data='validation', return_true=True,
                                               process_results=False)

                metrics = Metrics(test_y, p)
                val_score = getattr(metrics, val_metric)()
            else:
                if xy:
                    val_score = _model.cross_val_score(*data)
                else:
                    val_score = _model.cross_val_score(data=data)

            

            orig_val_score = val_score

            if metric_type != "min":
                val_score = 1.0 - val_score

            if not math.isfinite(val_score):
                val_score = DEFAULTS.get(val_metric, 1.0)

            print("{:<15} {:<20.5f} {:<20.5f}".format(self.iter, val_score, orig_val_score))
            self.iter += 1

            del _model
            gc.collect()

            return val_score

        optimizer = hpo.HyperOpt(
            self.algorithm,
            objective_fn=objective_fn,
            param_space=self.space,
            num_iterations=self.num_iterations,
            process_results=self.process_results,
            opt_path=f"results\\{PREFIX}"
        )

        optimizer.fit()

        return optimizer


class OptimizeHyperparameters(ModelOptimizerMixIn):

    def __init__(
            self,
            model,
            space,
            algorithm,
            num_iterations,
            process_results=False,
            **kwargs
    ):
        super().__init__(
            model=model,
            algorithm=algorithm,
            num_iterations=num_iterations,
            process_results=process_results,
            prefix="hpo"
        )

        self.space = space
        config = jsonize(model.config)
        model_config = config['model']
        # algo type is name of algorithm, e.g. xgboost, randomforest or layers
        self.algo_type = list(model_config.keys())[0]
        self.original_model = model._original_model_config

    def update(self, config, suggestions):
        # first update the model config parameters
        new_model_config = update_model_config(self.original_model['model'].copy(), suggestions)
        config['model'] = {self.algo_type: new_model_config}

        # now update hyperparameters which are not part of model config
        new_other_config = update_model_config(self.original_model['other'].copy(), suggestions)
        config.update(jsonize(new_other_config))

        return config


class OptimizeTransformations(ModelOptimizerMixIn):

    def __init__(
            self,
            model,
            categories,
            num_iterations=12,
            algorithm="bayes",
            include=None,
            exclude=None,
            append=None,
            process_results=False,
    ):
        super().__init__(
            model=model,
            num_iterations=num_iterations,
            algorithm=algorithm,
            process_results=process_results,
            prefix="trans_hpo",
        )

        self.space = make_space(self.model.input_features,
                                include=include,
                                exclude=exclude,
                                append=append,
                                categories=categories)

        self.input_features = model.input_features
        assert isinstance(self.input_features, list)
        self.output_features = model.output_features
        if isinstance(self.output_features, str):
            self.output_features = [self.output_features]
        assert len(self.output_features) == 1


    def update(self, config, suggestions):
        """updates `x_transformation` and `y_transformation` keys in config
        based upon `suggestions`."""
        transformations = []
        y_transformations = []

        for feature, method in suggestions.items():

            if method == "none":
                pass
            else:
                t = {"method": method, "features": [feature]}

                if method.startswith("log"):
                    t["treat_negatives"] = True
                    t["replace_zeros"] = True
                elif method in ["box-cox", "yeo-johnson", "power"]:
                    t["treat_negatives"] = True
                    t["replace_zeros"] = True
                elif method == "sqrt":
                    t['treat_negatives'] = True
                    t["replace_zeros"] = True

                if feature in self.input_features:
                    transformations.append(t)
                else:
                    y_transformations.append(t)

        # following parameters must be overwritten even if they were provided by the user.
        config['x_transformation'] = transformations
        config['y_transformation'] = y_transformations or None

        return


def make_space(
        input_features: list,
        categories: list,
        include: Union[str, list, dict] = None,
        exclude: Union[str, list] = None,
        append: dict = None,
) -> list:
    """
    Arguments:
        input_features :
        categories :
        include: the names of input features to include
        exclude: the name/names of input features to exclude
        append: the input features with custom candidate transformations
    """
    hpo = importlib.import_module("ai4water.hyperopt")
    Categorical = hpo.Categorical

    # although we need space as list at the end, it is better to create a dictionary of it
    # because manipulating dictionary is easier
    space = {
        name: Categorical(categories, name=name) for name in input_features
    }

    if include is not None:
        if isinstance(include, str):
            include = {include: categories}
        elif isinstance(include, list):
            _include = {}
            for feat in include:
                assert feat in input_features, f"{feat} is not in input_features but is used in include"
                if isinstance(feat, str):
                    _include[feat] = categories
                else:
                    assert isinstance(feat, dict) and len(feat) == 1
                    _include[list(feat.keys())[0]] = list(feat.values())[0]
            include = _include
        else:
            assert isinstance(include, dict) and len(include) == 1

        # since include is given, we will ignore default case when all features are considered
        space = {}
        for k, v in include.items():
            if not isinstance(v, Categorical):
                assert isinstance(v, list), f"space for {k} must be list but it is {v.__class__.__name__}"
                v = Categorical(v, name=k)
            space[k] = v

    if exclude is not None:
        if isinstance(exclude, str):
            exclude = [exclude]
        assert isinstance(exclude, list)
        for feat in exclude:
            space.pop(feat)

    if append is not None:
        assert isinstance(append, dict)
        for k, v in append.items():
            if not isinstance(v, Categorical):
                assert isinstance(v, list), f"space for {k} must be list but it is {v.__class__.__name__}"
                v = Categorical(v, name=k)
            space[k] = v

    return list(space.values())



import warnings
from typing import Union

import numpy as np
import pandas as pd

from ai4water.utils.utils import dateandtime_now, jsonize, deepcopy_dict_without_clone

from ._transformations import MinMaxScaler, PowerTransformer, QuantileTransformer, StandardScaler
from ._transformations import LogScaler, Log10Scaler, Log2Scaler, TanScaler, SqrtScaler, CumsumScaler
from ._transformations import FunctionTransformer, RobustScaler, MaxAbsScaler
from ._transformations import Center
from .utils import InvalidTransformation, TransformerNotFittedError, SP_METHODS


# TODO add logistic, tanh and more scalers.
# which transformation to use? Some related articles/posts
# https://scikit-learn.org/stable/modules/preprocessing.html
# http://www.faqs.org/faqs/ai-faq/neural-nets/part2/section-16.html
# https://data.library.virginia.edu/interpreting-log-transformations-in-a-linear-model/


class TransformationsContainer(object):

    def __init__(self):
        self.scalers = {}
        self.transforming_straight = True
        # self.nan_indices = None
        self.index = None


INITIATED_TRANSFORMERS = {
    'log': LogScaler(),
    'log2': Log2Scaler(),
    'log10': Log10Scaler(),
    'sqrt': SqrtScaler()
}

class Transformation(TransformationsContainer):
    """
    Applies transformation to tabular data. It is also possible to apply transformation
    on some selected features/columns of data. This class also performs some optional
    pre-processing on data before applying transformation on it.
    Any new transforming methods should define two methods one starting with
    `transform_with_` and `inverse_transofrm_with_`


    Currently following methods are available for transformation and inverse transformation

    Transformation methods

        - `minmax`
        - `maxabs`
        - `robust`
        - `power` same as yeo-johnson
        - `yeo-johnson`
        - `box-cox`
        - `zscore`    also known as standard scalers
        - `scale`    division by standard deviation
        - 'center'   by subtracting mean
        - `quantile`
        - `log`      natural logrithmic
        - `log10`    log with base 10
        - `log2`  log with base 2
        - `sqrt` square root
        - `tan`      tangent
        - `cumsum`   cummulative sum

    To transform a datafrmae using any of the above methods use

    Examples:
        >>> transformer = Transformation(method='zscore')
        >>> transformer.fit_transform(data=[1,2,3,5])

        or

        >>> transformer = Transformation(method='minmax')
        >>> normalized_df = transformer.fit_transform_with_minmax(data=pd.DataFrame([1,2,3]))

        >>> transformer = Transformation(method='minmax')
        >>> normalized_df, scaler_dict = transformer(data=pd.DataFrame([1,2,3]))

        or using one liner

        >>> normalized_df = Transformation(method='minmax',
        ...                       features=['a'])(data=pd.DataFrame([[1,2],[3,4], [5,6]],
        ...                                       columns=['a', 'b']))

    where `method` can be any of the above mentioned methods.

    Note
    ------
     `tan` and `cumsum` do not return original data upon inverse transformation.

    .. _google:
        https://developers.google.com/machine-learning/data-prep/transform/normalization
    """

    available_transformers = {
        "minmax": MinMaxScaler,
        "zscore": StandardScaler,
        "center": Center,
        "scale": StandardScaler,
        "robust": RobustScaler,
        "maxabs": MaxAbsScaler,
        "power": PowerTransformer,
        "yeo-johnson": PowerTransformer,
        "box-cox": PowerTransformer,
        "quantile": QuantileTransformer,
        "log": LogScaler,
        "log10": Log10Scaler,
        "log2": Log2Scaler,
        "sqrt": SqrtScaler,
        "tan": TanScaler,
        "cumsum": CumsumScaler
    }

    def __init__(self,
                 method: str = 'minmax',
                 features: list = None,
                 replace_zeros: bool = False,
                 replace_zeros_with: Union[str, int, float] = 1,
                 treat_negatives: bool = False,
                 **kwargs
                 ):
        """
        Arguments:
            method : method by which to transform and consequencly inversely
                transform the data. default is 'minmax'. see `Transformations.available_transformers`
                for full list.
            features : string or list of strings. Only applicable if `data` is
                dataframe. It defines the columns on which we want to apply transformation.
                The remaining columns will remain same/unchanged.
            replace_zeros : If true, then setting this argument to True will replace
                the zero values in data with some fixed value `replace_zeros_with`
                before transformation. The zero values will be put back at their
                places after transformation so this replacement/implacement is
                done only to avoid error during transformation for example during Box-Cox.
            replace_zeros_with : if replace_zeros is True, then this value will be used
                to replace zeros in dataframe before doing transformation. You can
                define the method with which to replace nans for exaple by setting
                this argument to 'mean' will replace zeros with 'mean' of the
                array/column which contains zeros. Allowed string values are
                'mean', 'max', 'min'. see_
            treat_negatives:
                If true, and if data contains negative values, then the absolute
                values of these negative values will be considered for transformation.
                For inverse transformation, the -ve sign is removed, to return the
                original data. This option is necessary for log, sqrt and box-cox
                transformations with -ve values in data.
            kwargs : any arguments which are to be provided to transformer on
                INTIALIZATION and not during transform or inverse transform

        Example:
            >>> from ai4water.preprocessing.transformations import Transformation
            >>> from ai4water.datasets import busan_beach
            >>> df = busan_beach()
            >>> inputs = ['tide_cm', 'wat_temp_c', 'sal_psu', 'air_temp_c', 'pcp_mm', 'pcp3_mm']
            >>> transformer = Transformation(method='minmax', features=['sal_psu', 'air_temp_c'])
            >>> new_data = transformer.fit_transform(df[inputs])

            Following shows how to apply log transformation on an array containing zeros
            by making use of the argument `replace_zeros`. The zeros in the input array
            will be replaced internally but will be inserted back afterwards.
            >>> from ai4water.preprocessing.transformations import Transformation
            >>> transformer = Transformation(method='log', replace_zeros=True)
            >>> transformed_data = transformer.fit_transform([1,2,3,0.0, 5, np.nan, 7])
            ... [0.0, 0.6931, 1.0986, 0.0, 1.609, None, 1.9459]
            >>> original_data = transformer.inverse_transform(data=transformed_data)

        .. _see:
            https://stats.stackexchange.com/a/222237/338323

        """
        super().__init__()

        if method not in self.available_transformers.keys():
            raise InvalidTransformation(method, list(self.available_transformers.keys()))

        self.method = method
        self.replace_zeros = replace_zeros
        self.replace_zeros_with = replace_zeros_with
        self.treat_negatives = treat_negatives

        self.features = features

        self.kwargs = kwargs
        self.transformed_features = None

    def __call__(self, data, what="fit_transform", return_key=False, **kwargs):
        """
        Calls the `fit_transform` and `inverse_transform` methods.
        """
        if what.startswith("fit"):
            self.transforming_straight = True

            return self.fit_transform(data, return_key=return_key, **kwargs)

        elif what.startswith("inv"):
            self.transforming_straight = False
            return self.inverse_transform(data, **kwargs)

        else:
            raise ValueError(f"The class Transformation can not be called with keyword argument 'what'={what}")

    def __getattr__(self, item):
        """
        Gets the attributes from underlying transformation modules.
        """
        if item.startswith('_'):
            return self.__getattribute__(item)
        elif item.startswith("fit_transform_with"):
            transformer = item.split('_')[-1]
            if transformer.lower() in list(self.available_transformers.keys()):
                self.method = transformer
                return self.fit_transform_with_sklearn

        elif item.startswith("inverse_transform_with"):
            transformer = item.split('_')[-1]
            if transformer.lower() in list(self.available_transformers.keys()):
                self.method = transformer
                return self.inverse_transform_with_sklearn
        else:
            raise AttributeError(f'Transformation has no attribute {item}')

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, x):
        if isinstance(x, pd.DataFrame):
            self._data = x
        else:
            assert isinstance(x, np.ndarray)
            xdf = pd.DataFrame(x, columns=['data'+str(i) for i in range(x.shape[1])])
            self._data = xdf

    @property
    def features(self):
        return self._features

    @features.setter
    def features(self, x):
        if x is not None:
            assert len(x) == len(set(x)), f"duplicated features are not allowed. Features are: {x}"
        self._features = x

    @property
    def transformed_features(self):
        return self._transformed_features

    @transformed_features.setter
    def transformed_features(self, x):
        self._transformed_features = x

    @property
    def num_features(self):
        return len(self.features)

    def get_scaler(self):

        return self.available_transformers[self.method.lower()]

    def pre_process_data(self, data):
        """Makes sure that data is dataframe and optionally replaces nans"""
        data = to_dataframe(data)

        # save the index if not already saved so that can be used later
        if self.index is None:
            self.index = data.index

        indices = {}
        if self.replace_zeros and self.transforming_straight:
            # instead of saving indices with column names, using column indices
            # because df.iloc[row_idx, col_idx] is better than df[col_name].iloc[row_idx]
            for col_idx, col in enumerate(data.columns):
                # find index containing 0s in corrent column of dataframe
                i = data.index[data[col] == 0.0]
                if len(i) > 0:
                    indices[col_idx] = i.values
                    if self.replace_zeros_with in ['mean', 'max', 'min']:
                        replace_with = float(getattr(np, 'nan' + self.replace_zeros_with)(data[col]))
                    else:
                        replace_with = self.replace_zeros_with
                    data.iloc[indices[col_idx], col_idx] = get_val(data[col], replace_with)

            #if self.zero_indices is None:
        self.zero_indices_ = indices

        indices = {}
        if self.treat_negatives:
            for col_idx, col in enumerate(data.columns):
                # find index containing negatives in corrent column of dataframe
                i = data.index[data[col] < 0.0]
                if len(i) > 0:
                    indices[col_idx] = i.values
                    # turn -ve values into positives
                    data[col] = data[col].abs()

        self.negative_indices_ = indices

        return data

    def post_process_data(self, data):
        """If nans/zeros were replaced with some value, put nans/zeros back."""
        data = data.copy()
        if self.replace_zeros:
            if hasattr(self, 'zero_indices_'):
                for col, idx in self.zero_indices_.items():
                    data.iloc[idx, col] = 0.0

        if self.treat_negatives:
            if hasattr(self, 'negative_indices_'):
                for col, idx in self.negative_indices_.items():
                    # invert the sign of those values which were originally -ve
                    data.iloc[idx, col] = -data.iloc[idx, col]
        return data

    def fit_transform_with_sklearn(
            self,
            data:Union[pd.DataFrame, np.ndarray],
            return_key=False,
            **kwargs):

        original_data = data.copy()
        to_transform = self.get_features(data)  # TODO, shouldn't kwargs go here as input?

        if self.method.lower() in ["log", "log10", "log2"]:

            if (to_transform.values < 0).any():
                raise InvalidValueError(self.method, "negative")

        _kwargs = {}
        if self.method == "scale":
            _kwargs['with_mean'] = False
        elif self.method in ['power', 'yeo-johnson', 'box-cox']:
            # a = np.array([87.52, 89.41, 89.4, 89.23, 89.92], dtype=np.float32).reshape(-1,1)
            # power transformers sometimes overflow with small data which causes inf error
            to_transform = to_transform.astype("float64")
            if self.method == "box-cox":
                _kwargs['method'] = "box-cox"

        for k,v in self.kwargs.items():
            if k in _kwargs:
                _kwargs.pop(k)

        scaler = self.get_scaler()(**_kwargs, **self.kwargs)

        data = scaler.fit_transform(to_transform, **kwargs)

        data = pd.DataFrame(data, columns=to_transform.columns)

        scaler = self.serialize_scaler(scaler, to_transform)

        data = self.maybe_insert_features(original_data, data)

        data = self.post_process_data(data)

        self.tr_data = data
        if return_key:
            return data, scaler
        return data

    def inverse_transform_with_sklearn(self,
                                       data,
                                       postprocess=True,
                                       without_fit=False,
                                       **kwargs):

        self.transforming_straight = False

        scaler = self.get_scaler_from_dict(**kwargs)

        original_data = data.copy()
        to_transform = self.get_features(data)

        if without_fit:
            data = scaler.inverse_transform_without_fit(to_transform)
        else:
            data = scaler.inverse_transform(to_transform)

        data = pd.DataFrame(data, columns=to_transform.columns)

        data = self.maybe_insert_features(original_data, data)

        if postprocess:
            data = self.post_process_data(data)

        return data

    def fit_transform(self, data, return_key=False, **kwargs):
        """
        Transforms the data

        Arguments:
            data : a dataframe or numpy ndarray or array like. The transformed or inversely
                transformed value will have the same type as data and will have
                the same index as data (in case data is dataframe). The shape of
                `data` is supposed to be (num_examples, num_features).
            return_key : whether to return the scaler or not. If True, then a
                tuple is returned which consists of transformed data and scaler itself.
            kwargs :
        """
        self.transforming_straight = True

        data = self.pre_process_data(data.copy())

        if self.features is None:
            self.features = list(data.columns)

        setattr(self, 'initial_shape_', data.shape)

        return getattr(self, "fit_transform_with_" + self.method)(data, return_key=return_key, **kwargs)

    def inverse_transform(self,
                          data,
                          postprocess=True,
                          without_fit=False,
                          **kwargs):
        """
        Inverse transforms the data.

        Parameters
        ---------
            data:
            postprocess : bool
            without_fit : bool
            kwargs : any of the folliwng keyword arguments

            - data: data on which to apply inverse transformation
            - key : key to fetch scaler
            - scaler : scaler to use for inverse transformation. If not given, then
                the available scaler is used.
        """
        self.transforming_straight = False

        # during transform, we convert to df even when input is list or np array
        # which inserts columns/features into data.
        data = to_dataframe(data)

        if 'key' in kwargs or 'scaler' in kwargs:
            pass
        elif len(self.scalers) == 1:
            kwargs['scaler'] = list(self.scalers.values())[0]
        elif self.method in SP_METHODS:
            kwargs['scaler'] = INITIATED_TRANSFORMERS[self.method]
            without_fit = True
        else:
            raise TransformerNotFittedError()

        if self.treat_negatives and hasattr(self, "negative_indices_"):
            for col, idx in self.negative_indices_.items():
                data.iloc[idx, col] = -data.iloc[idx, col]

        return getattr(self, "inverse_transform_with_" + self.method.lower())(
            data,
            postprocess=postprocess,
            without_fit=without_fit,
            **kwargs)

    def get_features(self, data) -> pd.DataFrame:

        if self.features is None:
            return data
        else:
            assert isinstance(self.features, list)
            return data[self.features]

    def serialize_scaler(self, scaler, to_transform):
        key = self.method + str(dateandtime_now())
        serialized_scaler = {
            "scaler": scaler,
            "shape": to_transform.shape,
            "key": key
        }
        self.scalers[key] = scaler

        return serialized_scaler

    def get_scaler_from_dict(self, **kwargs):
        if 'scaler' in kwargs:
            scaler = kwargs['scaler']
        elif 'key' in kwargs:
            scaler = self.scalers[kwargs['key']]
        else:
            raise TransformerNotFittedError()
        return scaler

    def maybe_insert_features(self, original_df, trans_df):

        if self.index is not None:
            # the data on which to perform inverse transformation has different
            # length so can't insert index, therefore make sure that negatives and
            # zeros were not treated beccause their indexes do not remain same now
            if len(trans_df) != len(self.index):
                assert len(self.zero_indices_) == 0
                assert len(self.negative_indices_) == 0
            else:
                trans_df.index = self.index

        num_features = len(original_df.columns)
        if len(trans_df.columns) != num_features:
            df = pd.DataFrame(index=self.index)
            for col in original_df.columns:  # features:
                if col in trans_df.columns:
                    _df = trans_df[col]
                else:
                    _df = original_df[col]

                df = pd.concat([df, _df], axis=1)
        else:
            df = trans_df

        assert df.shape == original_df.shape, f"shape changed from {original_df.shape} to {df.shape}"
        return df

    def config(self)->dict:
        """returns a dictionary which can be used to reconstruct `Transformation`
        class using `from_config`.
        Returns:
            a dictionary
        """
        assert len(self.scalers) > 0, f"Transformation is not fitted yet"

        assert len(self.scalers) == 1
        scaler = list(self.scalers.values())[0].config()

        return {
            "scaler": {self.method: scaler},
            "shape": self.initial_shape_,
            "method": self.method,
            "features": self.features,
            "replace_zeros": self.replace_zeros,
            "replace_zeros_with": self.replace_zeros_with,
            "treat_negatives": self.treat_negatives,
            "kwargs": self.kwargs,
            "negative_indices_": jsonize(self.negative_indices_),
            "zero_indices_": jsonize(self.zero_indices_)
        }


    @classmethod
    def from_config(
            cls,
            config:dict
    )-> "Transformation":
        """constructs the `Transformation` class from `config` which has
        already been fitted/transformed.
        Arguments:
            config:
                a dicionary which is the output of `config()` method.
        Returns:
            an instance of `Transformation` class.
        """
        config = deepcopy_dict_without_clone(config)

        shape = config.pop('shape')
        scaler = config.pop('scaler')
        negative_indices_ = config.pop("negative_indices_")
        zero_indices_ = config.pop("zero_indices_")

        assert len(scaler) == 1
        scaler_name = list(scaler.keys())[0]
        scaler_config = list(scaler.values())[0]

        transformer = cls(**config)

        # initiate the scalers
        sc_initiated = transformer.available_transformers[scaler_name].from_config(scaler_config)


        transformer.scalers = {scaler_name: sc_initiated}
        transformer.negative_indices_ = negative_indices_
        transformer.zero_indices_ = zero_indices_
        transformer.initial_shape_ = shape

        return transformer


def get_val(df: pd.DataFrame, method):

    if isinstance(method, str):
        if method.lower() == "mean":
            return df.mean()
        elif method.lower() == "max":
            return df.max()
        elif method.lower() == "min":
            return df.min()
    elif isinstance(method, int) or isinstance(method, float):
        return method
    else:
        raise ValueError(f"unknown method {method} to replace nan vlaues")


class InvalidValueError(Exception):
    def __init__(self, method, reason):
        self.method = method
        self.reason = reason

    def remedy(self):
        if self.reason == "NaN":
            return "Try setting 'replace_nans' to True"
        elif self.reason == "zero":
            return "Try setting 'replace_zeros' to True"
        elif self.reason == "negative":
            return "Try setting 'treat_negatives' to True"

    def __str__(self):
        return (f"""
Input data contains {self.reason} values so {self.method} transformation
can not be applied.
{self.remedy()}
""")


def to_dataframe(data)->pd.DataFrame:

    if isinstance(data, pd.DataFrame):
        data = data
    else:
        data = np.array(data)
        if data.ndim == 1:
            data = data.reshape(-1, 1)
        assert isinstance(data, np.ndarray)
        data = pd.DataFrame(data, columns=['data' + str(i) for i in range(data.shape[1])])
    return data



import numpy as np

from ._main import BaseModel
from ai4water.tf_attributes import ACTIVATION_LAYERS, LAYERS, tf, keras, tcn
from .nn_tools import get_add_call_args, get_call_args


class Model(BaseModel):

    """
    Model class with Functional API and inherits from `BaseModel`.

    For ML/non-Neural Network based models, there is no difference in functional
    or sub-clsasing api. For DL/NN-based models, this class implements functional
    api and differs from subclassing api in internal implementation of NN. This
    class is usefull, if you want to use the functional API of keras to build
    your own NN structure. In such as case you can construct your NN structure
    by overwriting `add_layers`. Another advantage of this class is that sometimes,
    model_subclsasing is not possible for example due to some bugs in tensorflow.
    In such a case this class can be used. Otherwise all the features of ai4water
    are available in this class as well.

    Example:
        >>>from ai4water.functional import Model
    """

    def __init__(self, *args, **kwargs):
        """
        Initializes and builds the NN/ML model.
        """
        self._go_up = True
        super().__init__(*args, **kwargs)

        def_KModel = None
        if keras is not None:
            def_KModel = keras.models.Model
        self.KModel = kwargs.get('KModel', def_KModel)

        self.build()

    @property
    def api(self):
        return 'functional'

    @property
    def KModel(self):
        """sets k_model.

        In case when we want to customize the model such as for implementing custom
        `train_step`, we can provide the customized model as input the this Model
        class
        """
        return self._k_model

    @KModel.setter
    def KModel(self, x):
        self._k_model = x

    @property
    def weights(self):
        """Returns names of weights in model."""
        _ws = []
        for w in self._model.weights:
            _ws.append(w.name)
        return _ws

    @property
    def layer_names(self):
        _all_layers = []
        if self.category == "ML":
            return None
        for layer in self._model.layers:
            _all_layers.append(layer.name)
        return _all_layers

    @property
    def num_input_layers(self) -> int:
        if self.category != "DL":
            return np.inf
        else:
            return len(self._model.inputs)

    @property
    def input_layer_names(self) -> list:

        return [lyr.name.split(':')[0] for lyr in self._model.inputs]

    @property
    def layers_out_shapes(self) -> dict:
        """ returns shapes of outputs from all layers in model as dictionary"""
        shapes = {}

        for lyr in self._model.layers:
            shapes[lyr.name] = lyr.output_shape

        return shapes

    @property
    def layers_in_shapes(self) -> dict:
        """ returns the shapes of inputs to all layers"""
        shapes = {}

        for lyr in self._model.layers:
            shapes[lyr.name] = lyr.input_shape

        return shapes

    @property
    def fit_fn(self):
        return self._model.fit

    @property
    def evaluate_fn(self):
        return self._model.evaluate

    @property
    def predict_fn(self):
        return self._model.predict

    def _get_dummy_input_shape(self):
        shape = ()
        if self.config['backend'] == 'tensorflow' and self.category == "DL":
            if isinstance(self.model_.inputs, list):
                if len(self.model_.inputs) == 1:
                    shape = self.model_.inputs[0].shape
                else:
                    shape = [inp.shape for inp in self.model_.inputs]

        return shape

    def first_layer_shape(self):
        """ instead of tuple, returning a list so that it can be moified if needed"""
        if self.num_input_layers > 1:
            shapes = {}
            for lyr in self._model.inputs:
                shapes[lyr.name] = lyr.shape
            return shapes
        shape = []
        for idx, d in enumerate(self._model.layers[0].input.shape):
            if int(tf.__version__[0]) == 1:
                if isinstance(d, tf.Dimension):  # for tf 1.x
                    d = d.value

            if idx == 0:  # the first dimension must remain undefined so that the user may define batch_size
                d = -1
            shape.append(d)
        return shape

    def add_layers(self, layers_config: dict, inputs=None):
        """
        Builds the NN from dictionary.

        Arguments:
            layers_config : wholse keys can be one of the following:
                `config`: `dict`/lambda, Every layer must contain initializing
                    arguments as `config` dictionary. The `config` dictionary
                    for every layer can contain `name` key and its value must be
                    `str` type. If `name` key  is not provided in the config,
                    the provided layer name will be used as its name e.g in following case
                        layers = {'LSTM': {'config': {'units': 16}}}
                    the name of `LSTM` layer will be `LSTM` while in follwoing case
                        layers = {'LSTM': {'config': {'units': 16, 'name': 'MyLSTM'}}}
                    the name of the lstm will be `MyLSTM`.
                `inputs`: str/list,  The calling arguments for the list. If `inputs`
                    key is missing for a layer, it will be supposed that either
                    this is an Input layer or it uses previous outputs as inputs.
                `outputs`: str/list  We can specifity the outputs from a layer
                    by using the `outputs` key. The value to `outputs` must be a
                    string or list of strings specifying the name of outputs from
                    current layer which can be used later in the mdoel.
                `call_args`: str/list  We can also specify additional call arguments
                    by `call_args` key. The value to `call_args` must be a string
                    or a list of strings.
            inputs : if None, it will be supposed the the `Input` layer either
                exists in `layers_config` or an Input layer will be created
                within this method before adding any other layer. If not None,
                then it must be in `Input` layer and the remaining NN architecture
                will be built as defined in `layers_config`. This can be handy
                when we want to use this method several times to build a complex
                or parallel NN structure. avoid `Input` in layer names.
        Returns:
            inputs :
            outputs :
        """
        lyr_cache = {}
        wrp_layer = None  # indicator for wrapper layers
        first_layer = True
        idx = 0

        for lyr, lyr_args in layers_config.items():
            idx += 0

            if callable(lyr) and hasattr(lyr, '__call__'):
                LAYERS[lyr.__name__] = lyr
                self.config['model']['layers'] = update_layers_config(layers_config, lyr)
                lyr = lyr.__name__

            lyr_config, lyr_inputs, named_outs, call_args = self.deconstruct_lyr_args(lyr, lyr_args)

            if callable(lyr) and not hasattr(lyr, '__call__'):
                lyr = "lambda"

            lyr_name, args, lyr_config, activation = self.check_lyr_config(lyr, lyr_config)

            # may be user has defined layers without input layer, in this case add Input layer as first layer
            if first_layer:
                if inputs is not None:  # This method was called by providing it inputs.
                    assert isinstance(inputs, tf.Tensor)
                    lyr_cache["Input"] = inputs
                    # since inputs have been defined, all the layers that will be added will be next to first layer
                    first_layer = False
                    layer_outputs = inputs
                    assign_dummy_name(layer_outputs, 'input')
                elif lyr_name != "Input":
                    if 'input_shape' in lyr_config:  # input_shape is given in the first layer so make input layer
                        layer_outputs = LAYERS["Input"](shape=lyr_config['input_shape'])
                        assign_dummy_name(layer_outputs, 'input')
                    else:
                        # for simple dense layer based models, lookback will not be used
                        def_shape = (self.num_ins,) if self.lookback == 1 else (self.lookback, self.num_ins)
                        layer_outputs = LAYERS["Input"](shape=def_shape)

                    # first layer is built so next iterations will not be for first layer
                    first_layer = False
                    # put the first layer in memory to be used for model compilation
                    lyr_cache["Input"] = layer_outputs
                    # add th layer which the user had specified as first layer

                    assign_dummy_name(layer_outputs, 'input')

            if lyr_inputs is None:  # The inputs to the layer have not been specified, so either it is an Input layer
                # or it uses the previous outputs as inputs
                if lyr_name == "Input":
                    # it is an Input layer, hence should not be called
                    layer_outputs = LAYERS[lyr_name](*args, **lyr_config)
                    assign_dummy_name(layer_outputs, 'input')
                else:
                    # it is executable and uses previous outputs as inputs
                    if lyr_name in ACTIVATION_LAYERS:
                        layer_outputs = ACTIVATION_LAYERS[lyr_name](name=lyr_config['name'])(layer_outputs)
                    elif lyr_name in ['TimeDistributed', 'Bidirectional']:
                        wrp_layer = LAYERS[lyr_name]
                        lyr_cache[lyr_name] = wrp_layer
                        continue
                    elif "LAMBDA" in lyr_name.upper():
                        # lyr_config is serialized lambda layer, which needs to be deserialized
                        # by default the lambda layer takes the previous output as input
                        # however when `call_args` are provided, they overwrite the layer_outputs
                        if call_args is not None:  # todo, add example in docs
                            layer_outputs = get_add_call_args(call_args, lyr_cache, lyr_config['name'])
                        layer_outputs = tf.keras.layers.deserialize(lyr_config)(layer_outputs)
                        # layers_config['lambda']['config'] still contails lambda, so we need to replace the python
                        # object (lambda) with the serialized version (lyr_config) so that it can be saved as json file.
                        layers_config[lyr]['config'] = lyr_config
                    else:
                        if wrp_layer is not None:
                            layer_outputs = wrp_layer(LAYERS[lyr_name](*args, **lyr_config))(layer_outputs)
                            wrp_layer = None
                        else:
                            add_args = get_add_call_args(call_args, lyr_cache, lyr_config['name'])
                            layer_initialized = LAYERS[lyr_name](*args, **lyr_config)
                            layer_outputs = layer_initialized(layer_outputs, **add_args)
                            self.get_and_set_attrs(layer_initialized)

            else:  # The inputs to this layer have been specified so they must exist in lyr_cache.
                # it is an executable
                if lyr_name in ACTIVATION_LAYERS:
                    call_args, add_args = get_call_args(lyr_inputs, lyr_cache, call_args, lyr_config['name'])
                    layer_outputs = ACTIVATION_LAYERS[lyr_name](name=lyr_config['name'])(call_args, **add_args)
                elif lyr_name in ['TimeDistributed', 'Bidirectional']:
                    wrp_layer = LAYERS[lyr_name]
                    lyr_cache[lyr_name] = wrp_layer
                    continue
                elif "LAMBDA" in lyr_name.upper():
                    call_args, add_args = get_call_args(lyr_inputs, lyr_cache, call_args, lyr_config['name'])
                    layer_outputs = tf.keras.layers.deserialize(lyr_config)(call_args)
                    layers_config[lyr]['config'] = lyr_config
                else:
                    if wrp_layer is not None:
                        call_args, add_args = get_call_args(lyr_inputs, lyr_cache, call_args, lyr_config['name'])
                        layer_outputs = wrp_layer(LAYERS[lyr_name](*args, **lyr_config))(call_args, **add_args)
                        wrp_layer = None
                    else:
                        call_args, add_args = get_call_args(lyr_inputs, lyr_cache, call_args, lyr_config['name'])
                        layer_initialized = LAYERS[lyr_name](*args, **lyr_config)
                        # todo, following conditioning is not good
                        # for concat layer inputs should be ([a,b,c]) instaed of (a,b,c)
                        if isinstance(lyr_inputs, list) and lyr_name != "Concatenate":
                            layer_outputs = layer_initialized(*call_args, **add_args)
                        else:
                            layer_outputs = layer_initialized(call_args, **add_args)
                        self.get_and_set_attrs(layer_initialized)

            if activation is not None:  # put the string back to dictionary to be saved in config file
                lyr_config['activation'] = activation

            if named_outs is not None:

                if isinstance(named_outs, list):
                    # this layer is returning more than one output
                    assert len(named_outs) == len(layer_outputs), "Layer {} is expected to return {} " \
                                                                  "outputs but it actually returns " \
                                                                  "{}".format(lyr_name, named_outs, layer_outputs)
                    for idx, out_name in enumerate(named_outs):
                        self.update_cache(lyr_cache, out_name, layer_outputs[idx])
                else:
                    # this layer returns just one output, TODO, this might be re
                    self.update_cache(lyr_cache, named_outs, layer_outputs)

            self.update_cache(lyr_cache, lyr_config['name'], layer_outputs)
            first_layer = False

        #layer_outputs = self.maybe_add_output_layer(layer_outputs, lyr_cache)

        inputs = []
        for k, v in lyr_cache.items():
            # since the model is not build yet and we have access to only output tensors of each list, this is probably
            # the only way to know that how many `Input` layers were encountered during the run of this method. Each
            # tensor (except TimeDistributed) has .op.inputs attribute,
            # which is empty if a tensor represents output of Input layer.
            if int(''.join(tf.__version__.split('.')[0:2]).ljust(3, '0')) < 240:
                if k != "TimeDistributed" and hasattr(v, 'op'):
                    if hasattr(v.op, 'inputs'):
                        _ins = v.op.inputs
                        if len(_ins) == 0:
                            inputs.append(v)
            # not sure if this is the proper way of checking if a layer receives an input or not!
            else:
                if hasattr(v, '__dummy_name'):
                    inputs.append(v)

        # for case when {Input -> Dense, Input_1}, this method wrongly makes Input_1 as output so in such case use
        # {Input_1, Input -> Dense }, thus it makes Dense as output and first 2 as inputs, so throwing warning
        if int(''.join(tf.__version__.split('.')[0:2]).ljust(3, '0')) < 240:
            if len(layer_outputs.op.inputs) < 1:
                print("Warning: the output is of Input tensor class type")
        else:
            if 'op' not in dir(layer_outputs):  # layer_outputs does not have `op`, which means it has no incoming node
                print("Warning: the output is of Input tensor class type")

        return inputs, layer_outputs

    def compile(self, model_inputs, outputs, **compile_args):

        k_model = self.KModel(inputs=model_inputs, outputs=outputs)

        k_model.compile(loss=self.loss(), optimizer=self.get_optimizer(), metrics=self.get_metrics(), **compile_args)

        if self.verbosity > 0:
            k_model.summary()

        self.plot_model(k_model)
        return k_model

    def build(self, input_shape=None):

        self.print_info()

        if self.category == "DL":
            if self.config.get('model', None) is None:
                lyrs = None
            else:
                lyrs = self.config['model']['layers']

            inputs, predictions = self.add_layers(lyrs)

            self._model = self.compile(inputs, predictions)

            self.info['model_parameters'] = int(self._model.count_params()) if self._model is not None else None

            if self.verbosity > 0 and self.config['model'] is not None:
                if 'tcn' in self.config['model']['layers']:
                    if int(''.join(tf.__version__.split('.')[0:2]).ljust(3, '0')) >= 250:
                        # tf >= 2.5 does not have _layers and tcn uses _layers
                        setattr(self._model, '_layers', self._model.layers)
                    tcn.tcn_full_summary(self._model, expand_residual_blocks=True)
        else:
            self.build_ml_model()

        if not getattr(self, 'from_check_point', False):
            # fit may fail so better to save config before as well. This will be overwritten once the fit is complete
            self.save_config()

        self.update_info()

        return

    def loss_name(self):
        if isinstance(self._model.loss, str):
            return self._model.loss
        elif hasattr(self._model.loss, 'name'):
            return self._model.loss.name
        else:
            return self._model.loss.__name__


def update_layers_config(layers_config, lyr):
    new_config = {}
    for k, v in layers_config.items():
        if k == lyr:
            new_config[lyr.__name__] = v
        else:
            new_config[k] = v
    return new_config


def assign_dummy_name(tensor, dummy_name):
    if isinstance(tensor, list):
        for t in tensor:
            setattr(t, '__dummy_name', dummy_name)
    else:
        setattr(tensor, '__dummy_name', dummy_name)




from .et_methods import ETBase
from .et_methods import Abtew
from .et_methods import Albrecht
from .et_methods import BlaneyCriddle
from .et_methods import BrutsaertStrickler
from .et_methods import Camargo
from .et_methods import Caprio
from .et_methods import ChapmanAustralia
from .et_methods import Copais
from .et_methods import Dalton
from .et_methods import DeBruinKeijman
from .et_methods import DoorenbosPruitt
from .et_methods import GrangerGray
from .et_methods import Hamon
from .et_methods import HargreavesSamani
from .et_methods import Haude
from .et_methods import JensenHaiseBasins
from .et_methods import Kharrufa
from .et_methods import Linacre
from .et_methods import Makkink
from .et_methods import Irmak
from .et_methods import MattShuttleworth
from .et_methods import McGuinnessBordne
from .et_methods import Penman
from .et_methods import PenPan
from .et_methods import PenmanMonteith
from .et_methods import PriestleyTaylor
from .et_methods import Romanenko
from .et_methods import SzilagyiJozsa
from .et_methods import Thornthwait
from .et_methods import Ritchie
from .et_methods import Turc
from .et_methods import Valiantzas




from ._shap import ShapExplainer
from ._lime import LimeExplainer
from ._permutation_importance import PermutationImportance
from ._partial_dependence import PartialDependencePlot

from .explainer_helper_funcs import explain_model
from .explainer_helper_funcs import explain_model_with_shap
from .explainer_helper_funcs import explain_model_with_lime




import warnings

from .dataset import DataSet, DataSetUnion, DataSetPipeline
from .transformations import Transformation, Transformations
from .imputation import Imputation
from .spatial_processing import MakeHRUs

class DataHandler(DataSet):
    def __init__(self, *args, **kwargs):
        warnings.warn("""
        DataHandler is deprecated and will be removed in future versions.
        Use `DataSet` class instead as follows
        from ai4water.preprocessing import DataSet
        """)
        super().__init__(*args, **kwargs)




try:
    import torch
except (ModuleNotFoundError, ImportError):
    torch = None

LAYERS, LOSSES, OPTIMIZERS = {}, {}, {}
Learner = None

if torch is not None:
    from .HARHN import HARHN
    from .pytorch_training import Learner
    from .imv_networks import IMVTensorLSTM, IMVFullLSTM
    from .pytorch_attributes import LAYERS, LOSSES, OPTIMIZERS



"""
SeqMetrics has been moved to its own dedicated repository. This will save us
many sleepless nights later.
"""

from ._main import Metrics
from ._rgr import RegressionMetrics
from ._cls import ClassificationMetrics
from .utils import plot_metrics




import warnings

import numpy as np
from sklearn import preprocessing
# from sklearn.metrics import hinge_loss
from sklearn.metrics import balanced_accuracy_score
import sklearn

from .utils import list_subclass_methods, msg
from ._main import Metrics

from ai4water.backend import get_attributes

CLS_METRICS = get_attributes(sklearn.metrics, '_classification', case_sensitive=True)


class ClassificationMetrics(Metrics):
    """Calculates classification metrics."""
    # todo add very major erro and major error
    warnings.warn(msg("ClassificationMetrics"), UserWarning)

    def __init__(self, *args, multiclass=False, **kwargs):
        self.multiclass = multiclass
        super().__init__(*args, metric_type='classification', **kwargs)
        self.true_labels = self._true_labels()
        self.true_logits = self._true_logits()
        self.pred_labels = self._pred_labels()
        self.pred_logits = self._pred_logits()

        self.all_methods = list_subclass_methods(ClassificationMetrics, True)
        # self.all_methods = [m for m in all_methods if not m.startswith('_')]

    def __getattr__(self, item):

        def func(**kwargs):  # because we want .f1_score() and not .f1_score
            return CLS_METRICS[item](self.true_labels, self.pred_labels, **kwargs)
        return func


    @staticmethod
    def _minimal() -> list:
        """some minimal and basic metrics"""
        return list_subclass_methods(ClassificationMetrics, True)

    @staticmethod
    def _hydro_metrics() -> list:
        """some minimal and basic metrics"""
        return list_subclass_methods(ClassificationMetrics, True)

    def _num_classes(self):
        return len(self._classes())

    def _classes(self):
        array = self.true_labels
        return np.unique(array[~np.isnan(array)])

    def _true_labels(self):
        """retuned array is 1d"""
        if self.multiclass:
            if self.true.size == len(self.true):
                return self.true.reshape(-1,1)
            return np.argmax(self.true.reshape(-1,1), axis=1)
        # it should be 1 dimensional
        assert self.true.size == len(self.true)
        return self.true.reshape(-1,)

    def _true_logits(self):
        """returned array is 2d"""
        if self.multiclass:
            return self.true
        lb = preprocessing.LabelBinarizer()
        return lb.fit_transform(self.true)

    def _pred_labels(self):
        """returns 1d"""
        if self.multiclass:
            if self.predicted.size == len(self.predicted):
                return self.predicted.reshape(-1,1)
            return np.argmax(self.predicted.reshape(-1,1), axis=1)
        lb = preprocessing.LabelBinarizer()
        lb.fit(self.true_labels)
        return lb.inverse_transform(self.predicted)

    def _pred_logits(self):
        """returned array is 2d"""
        if self.multiclass:
            return self.true
        # we can't do it
        return None

    def cross_entropy(self, epsilon=1e-12):
        """
        Computes cross entropy between targets (encoded as one-hot vectors)
        and predictions.
        Input: predictions (N, k) ndarray
               targets (N, k) ndarray
        Returns: scalar
        """
        predictions = np.clip(self.predicted, epsilon, 1. - epsilon)
        n = predictions.shape[0]
        ce = -np.sum(self.true * np.log(predictions + 1e-9)) / n
        return ce

    # def hinge_loss(self):
    #     """hinge loss using sklearn"""
    #     if self.pred_logits is not None:
    #         return hinge_loss(self.true_labels, self.pred_logits)
    #     return None

    def balanced_accuracy_score(self):
        return balanced_accuracy_score(self.true_labels, self.pred_labels)

    def accuracy(self, normalize=True):
        if normalize:
            return np.average(self.true_labels==self.pred_labels)
        return (self.true_labels==self.pred_labels).sum()




__all__ = ["ACTIVATION_LAYERS", "ACTIVATION_FNS", "LOSSES", "LAYERS", "OPTIMIZERS", "tcn"]

# it is supposed that tf is available
from .backend import get_attributes

try:
    import tensorflow as tf
except ModuleNotFoundError:
    tf = None


try:
    import tcn
except ModuleNotFoundError:
    tcn = None

LOSSES = {}
LAYERS = {}

if tcn is not None:
    LAYERS.update({"TCN": tcn.TCN if tcn is not None else None})

if tf is not None:
    import ai4water.utils.tf_losses as tf_losses
    from ai4water.models.tensorflow import NBeats
    import ai4water.models.tensorflow.attention_layers as attns
    from ai4water.models.tensorflow import TemporalFusionTransformer
    keras = tf.keras
    LOSSES.update({
        'nse': tf_losses.tf_nse,
        'kge': tf_losses.tf_kge,
    })
    LOSSES.update(get_attributes(aus=tf.keras, what='losses', case_sensitive=True))
else:
    NBeats, TemporalFusionTransformer, attns, keras = None, None, None, None

if tf is not None:
    LAYERS.update({"TemporalFusionTransformer": TemporalFusionTransformer})
    LAYERS.update({"TFT": TemporalFusionTransformer})
    LAYERS.update(get_attributes(aus=tf.keras, what='layers', case_sensitive=True))


if NBeats is not None:
    LAYERS.update({"NBeats": NBeats})

if attns is not None:
    LAYERS.update(get_attributes(aus=attns, what='attn_layers', case_sensitive=True))

try:
    from .models.tensorflow.private_layers import PrivateLayers
except (ModuleNotFoundError, ImportError):
    PrivateLayers = None

if PrivateLayers is not None:
    # add private layers to dictionary
    LAYERS.update(get_attributes(aus=PrivateLayers, what='layers', case_sensitive=True))

ACTIVATION_LAYERS = {
    # https://ai.stanford.edu/%7Eamaas/papers/relu_hybrid_icml2013_final.pdf
    'LeakyReLU': lambda name='softsign': keras.layers.LeakyReLU(),
    # https://arxiv.org/pdf/1502.01852v1.pdf
    'PReLU': lambda name='prelu': keras.layers.PReLU(name=name),
    'relu': lambda name='relu': keras.layers.Activation('relu', name=name),
    'tanh': lambda name='tanh': keras.layers.Activation('tanh', name=name),
    'ELU': lambda name='elu': keras.layers.ELU(name=name),
    'ThresholdedReLU': lambda name='ThresholdRelu': keras.layers.ThresholdedReLU(name=name),
    'selu': lambda name='selu': keras.layers.Activation("selu", name=name),
    'sigmoid': lambda name='sigmoid': keras.layers.Activation('sigmoid', name=name),
    'hardsigmoid': lambda name='HardSigmoid': keras.layers.Activation('hard_sigmoid', name=name),
    'crelu': lambda name='crelu': keras.layers.Activation(tf.nn.crelu, name=name),
    'relu6': lambda name='relu6': keras.layers.Activation(tf.nn.relu6, name=name),
    'softmax': lambda name='softmax': keras.layers.Activation(tf.nn.softmax, name=name),
    'softplus': lambda name='sofplus': keras.layers.Activation(tf.nn.softplus, name=name),
    'softsign': lambda name='softsign': keras.layers.Activation(tf.nn.softsign, name=name),
    "swish": lambda name='swish': keras.layers.Activation(tf.nn.swish, name=name),
}

ACTIVATION_FNS = {
    'relu': 'relu',  # keras.layers.Activation('relu', name=name),
    'tanh': 'tanh',
    'elu': 'elu',
    "hardsigmoid": 'hard_sigmoid',
    "linear": 'linear',

}

if tf is not None:
    ACTIVATION_FNS.update({
        'leakyrelu': tf.nn.leaky_relu,
        'crelu': tf.nn.crelu,
        'selu': tf.nn.selu,  # tf.keras.activations.selu, # https://arxiv.org/pdf/1706.02515.pdf
        'relu6': tf.nn.relu6,  # http://www.cs.utoronto.ca/%7Ekriz/conv-cifar10-aug2010.pdf
        'softmax': tf.nn.softmax,
        "softsign": tf.nn.softsign,
        "softplus": tf.nn.softplus,
        'sigmoid': tf.nn.sigmoid,
        "swish": tf.nn.swish,  # https://arxiv.org/pdf/1710.05941.pdf
    })

OPTIMIZERS = {}
if tf is not None:
    OPTIMIZERS.update(get_attributes(aus=tf.keras, what='optimizers', case_sensitive=True))
